/*
    Aufgabe4) Interfaces

    Gegeben sind die Schnittstellen Movable, Copyable und ShapeOnPlane. ShapeOnPlane und Movable beziehen sich auf
    geometrische Objekte in einem 2D-Koordinatensystem. Point ist großteils gegeben und repräsentiert einen
    2D-Punkt mit x- und y-Koordinaten. Ergänzen Sie die fehlenden Teile zu Point.
    Zusätzlich sind folgende Klassen gefragt:

    - Point ist ein Punkt in der Ebene und implementiert Movable.

    - Triangle repräsentiert ein Dreieck bestehend aus 3 Eckpunkten (Point). Ein Triangle-Objekt wird
      mit der Angabe von drei Point-Objekten erzeugt. Schreiben Sie einen entsprechenden Konstruktor.

    - Circle repräsentiert einen Kreis mit Mittelpunkt (Point) und radius (double). Ein Circle-Objekt
      wird mit der Angabe von einem Mittelpunkt (Point) und dem Radius (double) erzeugt. Schreiben
      Sie einen entsprechenden Konstruktor.

    - Triangle und Circle sollen ShapeOnPlane implementieren. Dies ist die gemeinsame Schnittstelle
      für geometrische Formen in der 2D-Ebene. Solche Objekte können verschoben werden (move)
      und haben einen Umfang (perimeter). Weiters können von diesen Objekten
      mit der parameterlosen Methode copy() unabhängige Kopien erzeugt werden. Nachfolgende
      Änderungen (z.B. durch move) des Originals haben keine Auswirkungen auf die Kopie und umgekehrt.
    - Triangle und Circle sollen über eine öffentliche parameterlose Methode toString() verfügen,
      die eine lesbare Repräsentation (als String) des Objekts zurückliefert. Zahlen sollen dabei mit
      zwei Nachkommastellen dargestellt werden. Siehe toString()-Methode in Point sowie die Testfälle
      in Aufgabe4.main(). Hinweis: toString() wird implizit aufgerufen, wenn ein Objekt vom Typ
      Triange, Circle, Point, ShapeOnPlane, etc. mit System.out.println() ausgegeben wird.
      Daher wird der Aufruf in den Beispielen in Aufgabe4.main() teilweise nicht explizit angegeben.

    [fortgeschritten]
    Definieren Sie ein Interface DoubleMatrix. Alle Objekte die DoubleMatrix sind, verfügen über eine
    parameterlose Methode toArray() die ein zwei-dimensionales Array mit double-Werten liefert. In jeder Zeile
    des Arrays stehen die Koordinaten eines Punktes, oder ein Array mit nur einem Element im Fall eines
    Skalars (zB. Radius). Point, Circle und Triangle sollen DoubleMatrix implementieren.
    Schreiben Sie weiters eine static-Methode print, die einen Parameter vom Typ DoubleMatrix
    hat und den Inhalt des double[][]-Arrays ausgibt. Die Form des Arrays und das Format der Ausgabe erkennen
    Sie an den Testfällen unter Aufgabe4.main().

    Sie können in dieser Aufgabe auch Methoden aus java.util.Arrays nutzen.

    Zusatzfragen:
    1. Wozu benötigt man Interfaces?
    2. Was bedeutet es, wenn ein Interface ein anderes Interface erweitert (extends)?
    3. Welche der folgenden Ausdrücke sind nach der Anweisung
       Movable point = new Point(1d,2d);
       gültig? Warum?

        point.distanceTo(new Point(0d,0d))
        point.move(0.5,1d)
        point.copy()
        point.toString()
*/

import java.util.Locale;

interface Movable {

    //shifts the object by the vector (deltaX, deltaY)
    void move(double deltaX, double deltaY);

}

interface Copyable {

    //returns an independent copy of an object
    //should be declared/overridden by every subtype (incl. abstract subclasses)
    Copyable copy();

}

interface ShapeOnPlane extends Movable, Copyable {

    //returns the perimeter of the shape
    double perimeter();
    ShapeOnPlane copy();

}

class Point implements Movable, Copyable /*TODO: add clause here*/ {

    private double x, y;

    public Point(double x, double y) {

        this.x = x;
        this.y = y;

    }

    public Point(Point p) {

        this.x = p.x;
        this.y = p.y;

    }

    public String toString() {

        return "(" + String.format("%.2f",x) + "," + String.format("%.2f",y) + ")";

    }

    public void move(double deltaX, double deltaY) {

        this.x += deltaX;
        this.y += deltaY;

    }

    public double distanceTo(Point p) {

        return Math.hypot(p.x - this.x, p.y - this.y);

    }

    public Point copy() {

        return (new Point(this.x, this.y)) ;

    }

}



class Triangle implements ShapeOnPlane /* TODO: add clause here */ {

    /* TODO: complete Triangle-class definition */
    private Point A, B, C;

    public Triangle(Point A, Point B, Point C) {
        this.A = A;
        this.B = B;
        this.C = C;
    }

    @Override
    public double perimeter() {
        //Umfang Triangle
        return A.distanceTo(B) + A.distanceTo(C) + B.distanceTo(C);
    }

    @Override
    public Triangle copy() {
        return (new Triangle(new Point(this.A.copy()), new Point(this.B.copy()), new Point(this.C.copy())));
        // alternativ(?): return new Triangle(this);
    }

    @Override
    public void move(double deltaX, double deltaY) {
        this.A.move(deltaX, deltaY);
        this.B.move(deltaX, deltaY);
        this.C.move(deltaX, deltaY);
    }

    public String toString() {
        return "[" + this.A.toString() + "," +
                this.B.toString() + "," +
                this.C.toString() + "]";
    }
}

class Circle implements ShapeOnPlane /* TODO: add clause here */ {

    /* TODO: complete Circle-class definition */
    Point x;
    double radius;

    public Circle(Point x, double radius) {
        this.x = x;
        this.radius = radius;
    }

    @Override
    public double perimeter() {
        //Umfang vom Kreis
        return 2 * Math.PI * radius;
    }

    @Override
    public Circle copy() {
        return (new Circle(new Point(this.x.copy()), this.radius));
    }

    @Override
    public void move(double deltaX, double deltaY) {
        this.x.move(deltaX, deltaY);
    }

    public String toString(){
        return "[" + this.x.toString() + "," + String.format("%.2f",this.radius) + "]";
    }
}



//just for testing
public class Aufgabe4 {

    public static void main(String[] args) {

        //Inhalt wird nicht bewertet
        Locale.setDefault(Locale.US);

        //entfernen Sie zum Testen die folgenden Kommentarzeichen.

        /*
        ShapeOnPlane shape = new Circle(new Point(2.5, -1.0), 1.0);
        System.out.println(shape.toString()); //[(2.50,-1.00),1.00]
        ShapeOnPlane copy = shape.copy();
        shape.move(1.0,1.5);
        System.out.println(shape); //[(3.50,0.50),1.00]
        System.out.println(copy); //[(2.50,-1.00),1.00]
        System.out.println(shape.perimeter()); //6.283185307179586
        System.out.println(copy.perimeter()); //6.283185307179586
        shape.move(1.0,1.5);
        System.out.println(shape); //[(4.50,2.00),1.00]
        shape = new Triangle(new Point(2.5, 1.0), new Point(-1.0,3.0), new Point(0d,0d));
        System.out.println(shape.perimeter()); //9.885988937884907
        System.out.println(shape); //[(2.50,1.00), (-1.00,3.00), (0.00,0.00)]
        */

        /*
        //folgende Testfälle funktionieren nur, wenn Sie die Aufgabe für
        //Fortgeschrittene gelöst haben.
        //[fortgeschritten]
        DoubleMatrix matrix = new Circle(new Point(2.5, -1.0), 1.0);
        print(matrix); //[[2.5, -1.0][1.0]]
        matrix = new Triangle(new Point(2.5, 1.0), new Point(-1.0,3.0), new Point(0d,0d));
        print(matrix); //[[2.5, 1.0][-1.0, 3.0][0.0, 0.0]]
        */

    }

    /* [fortgeschritten] TODO: remove comment tags
    //[fortgeschritten]
    //public static void print (DoubleMatrix dm) {

        // [fortgeschritten] TODO: add lines here

    }
    */

}